#ifndef _SPARSE_H_
#define _SPARSE_H_
/*
   --------------------------------------------------------------
   File sparse.h of ADOL-C version 1.8.0          as of Dec/03/98
   --------------------------------------------------------------
   All "Easy To Use" interfaces of SPARSE package

   Last changes:
        981203 olvo: untransposing reverse
        981130 olvo: newly created from adutils.h & adutilsc.h

   --------------------------------------------------------------
*/


/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/*                                                        PUBLIC INTERFACES */

#ifdef __cplusplus
/****************************************************************************/
/****************************************************************************/
/*                                                       Now the C++ THINGS */

/****************************************************************************/
/*                                           FORWARD MODE, overloaded calls */

/*--------------------------------------------------------------------------*/
/*  Bit pattern propagation call, d = 1, tight version                      */
/*                                                                          */
/* forward(tag, m, n, p, x[n], X[n][p], y[m], Y[m][p], mode) : intfov       */

int forward(short, int, int, int, double*, unsigned int**, 
                                  double*, unsigned int**, short =0);

/*--------------------------------------------------------------------------*/
/*  Bit pattern propagation call, d = 1, safe version (no x[.] and y[.])    */
/*                                                                          */
/* forward(tag, m, n, p, X[n][p], Y[m][p], mode) : intfov                   */

int forward(short, int, int, int, unsigned int**, unsigned int**, short =0);


/****************************************************************************/
/*                                           REVERSE MODE, overloaded calls */


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  Bit pattern propagation call, d = 0, tight & safe version               */ 
/*                                                                          */
/* reverse(tag, m, n, q, U[q][m], Z[q][n], mode) : intfov                   */

int reverse(short, int, int, int, unsigned int**, unsigned int**, short =0);


/****************************************************************************/
/****************************************************************************/
/*                                                         Now the C THINGS */
extern "C" {
#endif


/*--------------------------------------------------------------------------*/
/*                                                         jacobian pattern */
/* jac_pat(tag, m, n, argument, 
           &rb[1+m], &cb[1+n],  
           &crs[rb[0]][ crs[...][0] = non-zero independent blocks per row ],
           options[2])                                                      

   if (rb == NULL) each dependent variable will be considered 
                   as a block of dependent variables.
                   (rb is returned in order to keep compatible to further 
                    new features)
   if (cb == NULL) each independent variable will be considered 
                   as a block of independent variables.
                   (cb is returned in order to keep compatible to further
                    new features)                                           */
               
int jac_pat(short,int,int,double*,unsigned short**,unsigned short**,
            unsigned short***,int[]);


/****************************************************************************/
/*                                                               THAT'S ALL */
#ifdef __cplusplus
}
#endif

#endif






